from __future__ import annotations

import time
from typing import TYPE_CHECKING

import numpy as np

from . import turbo_optimizer_utils, turbo_utils
from .components import AcquisitionOptimizer, Surrogate
from .components.builder import (
    build_acquisition_optimizer,
    build_surrogate,
    build_trust_region,
)
from .config.candidate_rv import CandidateRV
from .strategies import OptimizationStrategy
from .types.appendable_array import AppendableArray
from .types.telemetry import Telemetry

if TYPE_CHECKING:
    from numpy.random import Generator

    from .config.optimizer_config import OptimizerConfig


class Optimizer:
    def __init__(
        self,
        *,
        bounds: np.ndarray,
        config: OptimizerConfig,
        rng: Generator,
        surrogate: Surrogate,
        acquisition_optimizer: AcquisitionOptimizer,
        strategy: OptimizationStrategy | None = None,
    ) -> None:
        self._config = config
        bounds = np.asarray(bounds, dtype=float)
        if bounds.ndim != 2 or bounds.shape[1] != 2:
            raise ValueError(f"bounds must be (d, 2), got {bounds.shape}")
        self._bounds = bounds
        self._num_dim = bounds.shape[0]
        self._rng = rng
        self._surrogate = surrogate
        self._acq_optimizer = acquisition_optimizer
        self._strategy = (
            strategy
            if strategy is not None
            else config.init.init_strategy.create_runtime_strategy(
                bounds=self._bounds, rng=self._rng, num_init=config.init.num_init
            )
        )
        self._tr_state = build_trust_region(
            config.trust_region,
            num_dim=self._num_dim,
            rng=rng,
            candidate_rv=config.candidate_rv,
        )
        self._trailing_obs = (
            None if config.trailing_obs is None else int(config.trailing_obs)
        )
        self._gp_num_steps = 50
        if self._trailing_obs is not None and self._trailing_obs <= 0:
            raise ValueError(f"trailing_obs must be > 0, got {self._trailing_obs}")
        self._x_obs = AppendableArray()
        self._y_obs = AppendableArray()
        self._yvar_obs = AppendableArray()
        self._y_tr_list: list[float] | list[list[float]] = []
        self._expects_yvar: bool | None = None
        self._dt_fit = 0.0
        self._dt_gen = 0.0
        self._dt_sel = 0.0
        self._dt_tell = 0.0
        self._sobol_seed_base = int(rng.integers(2**31 - 1))
        self._restart_generation = 0
        self._incumbent_idx: int | None = None
        self._incumbent_x_unit: np.ndarray | None = None
        self._incumbent_y_scalar: np.ndarray | None = None

    @property
    def tr_obs_count(self) -> int:
        return len(self._y_obs)

    @property
    def tr_length(self) -> float:
        return float(self._tr_state.length)

    def telemetry(self) -> Telemetry:
        return Telemetry(
            dt_fit=self._dt_fit,
            dt_gen=self._dt_gen,
            dt_sel=self._dt_sel,
            dt_tell=self._dt_tell,
        )

    @property
    def init_progress(self) -> tuple[int, int] | None:
        return self._strategy.init_progress()

    def ask(self, num_arms: int) -> np.ndarray:
        num_arms = int(num_arms)
        if num_arms <= 0:
            raise ValueError(num_arms)
        turbo_optimizer_utils.reset_timing(self)
        return self._strategy.ask(self, num_arms)

    def _ask_normal(self, num_arms: int, *, is_fallback: bool = False) -> np.ndarray:
        self._tr_state.validate_request(num_arms, is_fallback=is_fallback)
        self._maybe_resample_weights()
        x_center = self._incumbent_x_unit
        if x_center is None:
            if len(self._y_obs) == 0:
                raise RuntimeError("no observations")
            x_center = np.full(self._num_dim, 0.5)
        t0 = time.perf_counter()
        lengthscales = self._surrogate.lengthscales
        x_cand = self._generate_candidates(x_center, lengthscales, num_arms=num_arms)
        self._dt_gen = time.perf_counter() - t0
        t0 = time.perf_counter()
        selected = self._acq_optimizer.select(
            x_cand,
            num_arms,
            self._surrogate,
            self._rng,
            tr_state=self._tr_state,
        )
        self._dt_sel = time.perf_counter() - t0
        return turbo_utils.from_unit(selected, self._bounds)

    def _find_x_center(self, x_obs: np.ndarray, y_obs: np.ndarray) -> np.ndarray | None:
        return self._incumbent_x_unit

    def _maybe_resample_weights(self) -> None:
        from .config.rescalarize import Rescalarize

        if hasattr(self._tr_state, "rescalarize"):
            if self._tr_state.rescalarize == Rescalarize.ON_PROPOSE:
                self._tr_state.resample_weights(self._rng)

    def _generate_candidates(
        self,
        x_center: np.ndarray,
        lengthscales: np.ndarray | None,
        *,
        num_arms: int,
    ) -> np.ndarray:
        if lengthscales is not None:
            lengthscales = np.asarray(lengthscales, dtype=float).reshape(-1)
            if not np.all(np.isfinite(lengthscales)):
                raise ValueError("lengthscales must be finite")
        num_candidates = int(
            self._config.candidates.num_candidates(
                num_dim=self._num_dim, num_arms=num_arms
            )
        )
        if num_candidates <= 0:
            raise ValueError(num_candidates)
        candidate_rv = self._config.candidate_rv
        if candidate_rv == CandidateRV.SOBOL:
            from scipy.stats import qmc

            sobol_seed = turbo_optimizer_utils.sobol_seed_for_state(
                self._sobol_seed_base,
                restart_generation=self._restart_generation,
                n_obs=len(self._x_obs),
                num_arms=num_arms,
            )
            sobol_engine = qmc.Sobol(d=self._num_dim, scramble=True, seed=sobol_seed)
        else:
            sobol_engine = None
        if getattr(self._tr_state, "uses_custom_candidate_gen", False):
            return self._tr_state.generate_candidates(
                x_center,
                lengthscales,
                num_candidates,
                rng=self._rng,
                sobol_engine=sobol_engine,
                raasp_driver=self._config.raasp_driver,
                num_pert=20,
            )
        return turbo_utils.generate_tr_candidates(
            self._tr_state.compute_bounds_1d,
            x_center,
            lengthscales,
            num_candidates,
            rng=self._rng,
            candidate_rv=candidate_rv,
            sobol_engine=sobol_engine,
            raasp_driver=self._config.raasp_driver,
            num_pert=20,
        )

    def _validate_tell_inputs(
        self, x: np.ndarray, y: np.ndarray, y_var: np.ndarray | None
    ) -> turbo_optimizer_utils.TellInputs:
        inputs = turbo_optimizer_utils.validate_tell_inputs(x, y, y_var, self._num_dim)
        tr_num_metrics = getattr(self._tr_state, "num_metrics", 1)
        if inputs.num_metrics != tr_num_metrics:
            raise ValueError(
                f"y has {inputs.num_metrics} metrics but trust region expects {tr_num_metrics}"
            )
        if self._expects_yvar is None:
            self._expects_yvar = inputs.y_var is not None
        if (inputs.y_var is not None) != bool(self._expects_yvar):
            raise ValueError(
                f"y_var must be {'provided' if self._expects_yvar else 'omitted'} on every tell()"
            )
        return inputs

    def _update_incumbent(self) -> None:
        if len(self._y_obs) == 0:
            self._incumbent_idx, self._incumbent_x_unit, self._incumbent_y_scalar = (
                None,
                None,
                None,
            )
            return
        x_obs, y_obs = self._x_obs.view(), self._y_obs.view()
        candidate_indices = self._surrogate.get_incumbent_candidate_indices(y_obs)
        x_cand, y_cand = x_obs[candidate_indices], y_obs[candidate_indices]
        mu_cand, noise_aware = None, False
        if hasattr(self._tr_state, "incumbent_selector"):
            noise_aware = getattr(
                self._tr_state.incumbent_selector, "noise_aware", False
            )
        elif hasattr(self._tr_state, "config"):
            noise_aware = getattr(self._tr_state.config, "noise_aware", False)

        if noise_aware:
            try:
                mu_cand = self._surrogate.predict(x_cand).mu
            except RuntimeError:
                mu_cand = None

        idx_in_cand = self._tr_state.get_incumbent_index(y_cand, self._rng, mu=mu_cand)
        self._incumbent_idx = int(candidate_indices[idx_in_cand])
        self._incumbent_x_unit = x_obs[self._incumbent_idx]
        self._incumbent_y_scalar = (
            mu_cand[idx_in_cand : idx_in_cand + 1]
            if noise_aware and mu_cand is not None
            else y_cand[idx_in_cand : idx_in_cand + 1]
        ).copy()

    def _trim_trailing_obs(self) -> None:
        incumbent_indices = np.array([self._incumbent_idx], dtype=int)
        obs = turbo_optimizer_utils.trim_trailing_observations(
            self._x_obs.view().tolist(),
            self._y_obs.view().tolist(),
            self._y_tr_list,
            self._yvar_obs.view().tolist() if len(self._yvar_obs) > 0 else [],
            trailing_obs=self._trailing_obs,
            incumbent_indices=incumbent_indices,
        )
        self._x_obs = AppendableArray()
        for x in obs.x_obs:
            self._x_obs.append(np.array(x))
        self._y_obs = AppendableArray()
        for y in obs.y_obs:
            self._y_obs.append(np.array(y))
        self._yvar_obs = AppendableArray()
        if obs.yvar_obs:
            for yvar in obs.yvar_obs:
                self._yvar_obs.append(np.array(yvar))
        self._y_tr_list = obs.y_tr

    def _update_best_value_if_needed(self) -> None:
        pass

    def tell(
        self, x: np.ndarray, y: np.ndarray, y_var: np.ndarray | None = None
    ) -> np.ndarray:
        with turbo_utils.record_duration(
            lambda dt: setattr(self, "_dt_tell", float(dt))
        ):
            inputs = self._validate_tell_inputs(x, y, y_var)
            if inputs.x.shape[0] == 0:
                return (
                    np.array([], dtype=float)
                    if inputs.num_metrics == 1
                    else np.empty((0, inputs.num_metrics), dtype=float)
                )
            x_unit = turbo_utils.to_unit(inputs.x, self._bounds)
            for i in range(inputs.x.shape[0]):
                self._x_obs.append(x_unit[i])
                self._y_obs.append(inputs.y[i])
                if inputs.y_var is not None:
                    self._yvar_obs.append(inputs.y_var[i])
            return self._strategy.tell(self, inputs, x_unit=x_unit)


def create_optimizer(
    *,
    bounds: np.ndarray,
    config: OptimizerConfig,
    rng: Generator,
) -> Optimizer:
    surrogate = build_surrogate(config.surrogate)
    base_acq_optimizer = build_acquisition_optimizer(config.acquisition)

    from .components.acquisition import (
        HnRAcqOptimizer,
        ThompsonAcqOptimizer,
        UCBAcqOptimizer,
    )
    from .config.acquisition import HnROptimizerConfig

    if isinstance(config.acq_optimizer, HnROptimizerConfig):
        if isinstance(base_acq_optimizer, (ThompsonAcqOptimizer, UCBAcqOptimizer)):
            acq_optimizer = HnRAcqOptimizer(base_acq_optimizer)
        else:
            raise ValueError(
                f"HnR not supported with {type(base_acq_optimizer).__name__}"
            )
    else:
        acq_optimizer = base_acq_optimizer

    return Optimizer(
        bounds=bounds,
        config=config,
        rng=rng,
        surrogate=surrogate,
        acquisition_optimizer=acq_optimizer,
    )



from __future__ import annotations
import numpy as np
from .candidates import Candidates
from .conditional_posterior_draw_internals import ConditionalPosteriorDrawInternals
from .enn_like_protocol import ENNLike
from .enn_params import ENNParams, PosteriorFlags
from .neighbors import Neighbors

_ENNLike = ENNLike
_Candidates = Candidates
_Neighbors = Neighbors


def _pairwise_sq_l2(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    a = np.asarray(a, dtype=float)
    b = np.asarray(b, dtype=float)
    aa = np.sum(a * a, axis=1, keepdims=True)
    bb = np.sum(b * b, axis=1, keepdims=True).T
    dist2 = aa + bb - 2.0 * (a @ b.T)
    return np.maximum(dist2, 0.0)


def _validate_x(enn: ENNLike, x: np.ndarray) -> np.ndarray:
    x = np.asarray(x, dtype=float)
    if x.ndim != 2 or x.shape[1] != enn._num_dim:
        raise ValueError(x.shape)
    return x


def _validate_whatif(
    enn: ENNLike, x_whatif: np.ndarray, y_whatif: np.ndarray
) -> tuple[np.ndarray, np.ndarray]:
    x_whatif = np.asarray(x_whatif, dtype=float)
    y_whatif = np.asarray(y_whatif, dtype=float)
    if x_whatif.ndim != 2 or x_whatif.shape[1] != enn._num_dim:
        raise ValueError(x_whatif.shape)
    if y_whatif.ndim != 2 or y_whatif.shape[1] != enn._num_metrics:
        raise ValueError(y_whatif.shape)
    if x_whatif.shape[0] != y_whatif.shape[0]:
        raise ValueError((x_whatif.shape, y_whatif.shape))
    return x_whatif, y_whatif


def _scale_x_if_needed(enn: ENNLike, x: np.ndarray) -> np.ndarray:
    return x / enn._x_scale if enn._scale_x else x


def _compute_total_n(enn: ENNLike, num_whatif: int, flags: PosteriorFlags) -> int:
    total_n = len(enn) + int(num_whatif)
    if flags.exclude_nearest and total_n <= 1:
        raise ValueError(total_n)
    return total_n


def _compute_search_k(params: ENNParams, flags: PosteriorFlags, total_n: int) -> int:
    return int(
        min(params.k_num_neighbors + (1 if flags.exclude_nearest else 0), total_n)
    )


def _get_train_candidates(enn: ENNLike, x: np.ndarray, *, search_k: int) -> Candidates:
    batch_size = x.shape[0]
    if len(enn) == 0 or search_k == 0:
        return Candidates(
            dist2=np.zeros((batch_size, 0), dtype=float),
            ids=np.zeros((batch_size, 0), dtype=int),
            y=np.zeros((batch_size, 0, enn._num_metrics), dtype=float),
            yvar=(
                np.zeros((batch_size, 0, enn._num_metrics), dtype=float)
                if enn._train_yvar is not None
                else None
            ),
        )
    train_search_k = int(min(search_k, len(enn)))
    dist2_train, idx_train = enn._enn_index.search(
        x, search_k=train_search_k, exclude_nearest=False
    )
    y_train = enn._train_y[idx_train]
    yvar_train = enn._train_yvar[idx_train] if enn._train_yvar is not None else None
    return Candidates(dist2=dist2_train, ids=idx_train, y=y_train, yvar=yvar_train)


def _get_whatif_candidates(
    enn: ENNLike,
    x: np.ndarray,
    x_whatif: np.ndarray,
    y_whatif: np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    x_scaled = _scale_x_if_needed(enn, x)
    x_whatif_scaled = _scale_x_if_needed(enn, x_whatif)
    dist2_whatif = _pairwise_sq_l2(x_scaled, x_whatif_scaled)
    batch_size = x.shape[0]
    y_whatif_batched = np.broadcast_to(
        y_whatif[np.newaxis, :, :], (batch_size, y_whatif.shape[0], y_whatif.shape[1])
    )
    return dist2_whatif, y_whatif_batched


_WhatifCandidateTuple = tuple[np.ndarray, np.ndarray, np.ndarray]


def _merge_candidates(
    enn: ENNLike,
    *,
    train: Candidates,
    whatif: _WhatifCandidateTuple,
) -> Candidates:
    dist2_whatif, ids_whatif, y_whatif_batched = whatif
    dist2_all = np.concatenate([train.dist2, dist2_whatif], axis=1)
    ids_all = np.concatenate([train.ids, ids_whatif], axis=1)
    y_all = np.concatenate([train.y, y_whatif_batched], axis=1)
    if train.yvar is None:
        return Candidates(dist2=dist2_all, ids=ids_all, y=y_all, yvar=None)
    batch_size = dist2_all.shape[0]
    num_whatif = dist2_whatif.shape[1]
    yvar_whatif = np.zeros((batch_size, num_whatif, enn._num_metrics))
    yvar_all = np.concatenate([train.yvar, yvar_whatif], axis=1)
    return Candidates(dist2=dist2_all, ids=ids_all, y=y_all, yvar=yvar_all)


def _select_sorted_candidates(dist2_all: np.ndarray, *, search_k: int) -> np.ndarray:
    batch_size, num_candidates = dist2_all.shape
    if search_k < num_candidates:
        sel = np.argpartition(dist2_all, kth=search_k - 1, axis=1)[:, :search_k]
    else:
        sel = np.broadcast_to(np.arange(num_candidates), (batch_size, num_candidates))
    sel_dist2 = np.take_along_axis(dist2_all, sel, axis=1)
    sel_order = np.argsort(sel_dist2, axis=1)
    return np.take_along_axis(sel, sel_order, axis=1)


def _take_along_axis_3d(a: np.ndarray, idx_2d: np.ndarray) -> np.ndarray:
    return np.take_along_axis(a, idx_2d[:, :, np.newaxis], axis=1)


def _make_empty_normal(enn: ENNLike, batch_size: int):
    from .enn_normal import ENNNormal

    internals = enn._empty_posterior_internals(batch_size)
    return ENNNormal(internals.mu, internals.se)


def _build_candidates(
    enn: ENNLike,
    x: np.ndarray,
    x_whatif: np.ndarray,
    y_whatif: np.ndarray,
    *,
    search_k: int,
) -> Candidates:
    train_candidates = _get_train_candidates(enn, x, search_k=search_k)
    dist2_whatif, y_whatif_batched = _get_whatif_candidates(enn, x, x_whatif, y_whatif)
    n_train = int(len(enn))
    ids_whatif = np.broadcast_to(
        n_train + np.arange(x_whatif.shape[0], dtype=int), dist2_whatif.shape
    )
    return _merge_candidates(
        enn,
        train=train_candidates,
        whatif=(dist2_whatif, ids_whatif, y_whatif_batched),
    )


def _select_effective_neighbors(
    candidates: Candidates,
    *,
    search_k: int,
    k: int,
    exclude_nearest: bool,
) -> Neighbors | None:
    sel = _select_sorted_candidates(candidates.dist2, search_k=search_k)
    if exclude_nearest:
        sel = sel[:, 1:]
    sel = sel[:, : int(min(k, sel.shape[1]))]
    if sel.shape[1] == 0:
        return None
    dist2s = np.take_along_axis(candidates.dist2, sel, axis=1)
    ids = np.take_along_axis(candidates.ids, sel, axis=1)
    y_neighbors = _take_along_axis_3d(candidates.y, sel)
    yvar_neighbors = (
        _take_along_axis_3d(candidates.yvar, sel)
        if candidates.yvar is not None
        else None
    )
    return Neighbors(dist2=dist2s, ids=ids, y=y_neighbors, yvar=yvar_neighbors)


def _compute_mu_se(
    enn: ENNLike,
    neighbors: Neighbors,
    *,
    params: ENNParams,
    flags: PosteriorFlags,
    y_scale: np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    stats = enn._compute_weighted_stats(
        neighbors.dist2,
        neighbors.y,
        yvar_neighbors=neighbors.yvar,
        params=params,
        observation_noise=flags.observation_noise,
        y_scale=y_scale,
    )
    return stats.mu, stats.se


def _compute_draw_internals(
    enn: ENNLike,
    neighbors: Neighbors,
    *,
    params: ENNParams,
    flags: PosteriorFlags,
    y_scale: np.ndarray,
) -> ConditionalPosteriorDrawInternals:
    stats = enn._compute_weighted_stats(
        neighbors.dist2,
        neighbors.y,
        yvar_neighbors=neighbors.yvar,
        params=params,
        observation_noise=flags.observation_noise,
        y_scale=y_scale,
    )
    return ConditionalPosteriorDrawInternals(
        idx=neighbors.ids.astype(int, copy=False),
        w_normalized=stats.w_normalized,
        l2=stats.l2,
        mu=stats.mu,
        se=stats.se,
    )


def _conditional_neighbors_nonempty_whatif(
    enn: ENNLike,
    x_whatif: np.ndarray,
    y_whatif: np.ndarray,
    x: np.ndarray,
    *,
    params: ENNParams,
    flags: PosteriorFlags,
) -> tuple[int, int, Neighbors | None]:
    batch_size = x.shape[0]
    search_k = _compute_search_k(
        params, flags, _compute_total_n(enn, x_whatif.shape[0], flags)
    )
    if search_k == 0:
        return batch_size, search_k, None
    candidates = _build_candidates(enn, x, x_whatif, y_whatif, search_k=search_k)
    neighbors = _select_effective_neighbors(
        candidates,
        search_k=search_k,
        k=params.k_num_neighbors,
        exclude_nearest=flags.exclude_nearest,
    )
    return batch_size, search_k, neighbors


def _compute_conditional_posterior_impl(
    enn: ENNLike,
    x_whatif: np.ndarray,
    y_whatif: np.ndarray,
    x: np.ndarray,
    *,
    params: ENNParams,
    flags: PosteriorFlags,
    y_scale: np.ndarray,
):
    from .enn_normal import ENNNormal

    x = _validate_x(enn, x)
    x_whatif, y_whatif = _validate_whatif(enn, x_whatif, y_whatif)
    if x_whatif.shape[0] == 0:
        return enn.posterior(x, params=params, flags=flags)
    batch_size, search_k, neighbors = _conditional_neighbors_nonempty_whatif(
        enn, x_whatif, y_whatif, x, params=params, flags=flags
    )
    if search_k == 0 or neighbors is None:
        return _make_empty_normal(enn, batch_size)
    mu, se = _compute_mu_se(enn, neighbors, params=params, flags=flags, y_scale=y_scale)
    return ENNNormal(mu, se)


def compute_conditional_posterior(
    enn: ENNLike,
    x_whatif: np.ndarray,
    y_whatif: np.ndarray,
    x: np.ndarray,
    *,
    params: ENNParams,
    flags: PosteriorFlags,
    y_scale: np.ndarray,
):
    return _compute_conditional_posterior_impl(
        enn, x_whatif, y_whatif, x, params=params, flags=flags, y_scale=y_scale
    )


def compute_conditional_posterior_draw_internals(
    enn: ENNLike,
    x_whatif: np.ndarray,
    y_whatif: np.ndarray,
    x: np.ndarray,
    *,
    params: ENNParams,
    flags: PosteriorFlags,
    y_scale: np.ndarray,
) -> ConditionalPosteriorDrawInternals:
    x = _validate_x(enn, x)
    x_whatif, y_whatif = _validate_whatif(enn, x_whatif, y_whatif)
    if x_whatif.shape[0] == 0:
        raise ValueError("x_whatif must be non-empty for conditional draw internals")
    batch_size, search_k, neighbors = _conditional_neighbors_nonempty_whatif(
        enn, x_whatif, y_whatif, x, params=params, flags=flags
    )
    if search_k == 0 or neighbors is None:
        empty_internals = enn._empty_posterior_internals(batch_size)
        return ConditionalPosteriorDrawInternals(
            idx=empty_internals.idx,
            w_normalized=empty_internals.w_normalized,
            l2=empty_internals.l2,
            mu=empty_internals.mu,
            se=empty_internals.se,
        )
    return _compute_draw_internals(
        enn, neighbors, params=params, flags=flags, y_scale=y_scale
    )



from .lhd_only_strategy import LHDOnlyStrategy
from .optimization_strategy import OptimizationStrategy
from .turbo_hybrid_strategy import TurboHybridStrategy

__all__ = [
    "LHDOnlyStrategy",
    "OptimizationStrategy",
    "TurboHybridStrategy",
]



from __future__ import annotations
from typing import Any


def validate_optimizer_config(cfg: Any) -> None:
    from .acquisition import (
        DrawAcquisitionConfig,
        HnROptimizerConfig,
        NDSOptimizerConfig,
        ParetoAcquisitionConfig,
        UCBAcquisitionConfig,
    )
    from .init_strategies import LHDOnlyInit
    from .surrogate import GPSurrogateConfig, NoSurrogateConfig

    if isinstance(cfg.init.init_strategy, LHDOnlyInit):
        if not isinstance(cfg.surrogate, NoSurrogateConfig):
            raise ValueError(
                "init_strategy='lhd_only' requires NoSurrogateConfig surrogate"
            )
    if isinstance(cfg.surrogate, NoSurrogateConfig):
        if isinstance(cfg.acquisition, DrawAcquisitionConfig):
            raise ValueError(
                "DrawAcquisitionConfig (Thompson sampling) requires a surrogate. "
                "NoSurrogateConfig is not compatible with DrawAcquisitionConfig."
            )
        if isinstance(cfg.acquisition, UCBAcquisitionConfig):
            raise ValueError(
                "UCBAcquisitionConfig requires a surrogate. "
                "NoSurrogateConfig is not compatible with UCBAcquisitionConfig."
            )
    if isinstance(cfg.acquisition, ParetoAcquisitionConfig):
        if not isinstance(cfg.acq_optimizer, NDSOptimizerConfig):
            raise ValueError("ParetoAcquisitionConfig requires NDSOptimizerConfig")
    if isinstance(cfg.acq_optimizer, HnROptimizerConfig):
        if isinstance(cfg.acquisition, ParetoAcquisitionConfig):
            raise ValueError(
                "HnROptimizerConfig is not compatible with ParetoAcquisitionConfig"
            )
        if isinstance(cfg.surrogate, GPSurrogateConfig) and isinstance(
            cfg.acquisition, DrawAcquisitionConfig
        ):
            raise NotImplementedError(
                "GP surrogate with DrawAcquisitionConfig and HnROptimizerConfig is not yet implemented"
            )



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
from .posterior_result import PosteriorResult
from .surrogate_result import SurrogateResult

if TYPE_CHECKING:
    from numpy.random import Generator


class NoSurrogate:
    def __init__(self) -> None:
        self._x_obs: np.ndarray | None = None
        self._y_obs: np.ndarray | None = None

    @property
    def lengthscales(self) -> np.ndarray | None:
        return getattr(self, "_lengthscales", None)

    def fit(
        self,
        x_obs: np.ndarray,
        y_obs: np.ndarray,
        y_var: np.ndarray | None = None,
        *,
        num_steps: int = 0,
        rng: Generator | None = None,
    ) -> SurrogateResult:
        self._x_obs = np.asarray(x_obs, dtype=float)
        self._y_obs = np.asarray(y_obs, dtype=float)
        if self._y_obs.ndim == 1:
            self._y_obs = self._y_obs.reshape(-1, 1)
        return SurrogateResult(model=None, lengthscales=None)

    def predict(self, x: np.ndarray) -> PosteriorResult:
        if self._x_obs is None or self._y_obs is None:
            raise RuntimeError("NoSurrogate.predict requires fit() first")
        x = np.asarray(x, dtype=float)
        if np.array_equal(x, self._x_obs):
            return PosteriorResult(mu=self._y_obs.copy(), sigma=None)
        raise RuntimeError("NoSurrogate.predict only works for training points")

    def get_incumbent_candidate_indices(self, y_obs: np.ndarray) -> np.ndarray:
        return np.arange(len(y_obs), dtype=int)

    def sample(self, x: np.ndarray, num_samples: int, rng: Generator) -> np.ndarray:
        n = len(x)
        num_metrics = self._y_obs.shape[1] if hasattr(self, "_y_obs") else 1
        return rng.standard_normal((num_samples, n, num_metrics))



from __future__ import annotations
import pytest
from enn.turbo.config import (
    default_num_candidates,
    const_num_candidates,
    CandidateGenConfig,
)
from enn.turbo.config import CandidateRV


def test_default_num_candidates():
    assert default_num_candidates(num_dim=10, num_arms=4) == 1000
    assert default_num_candidates(num_dim=100, num_arms=4) == 5000


def test_const_num_candidates():
    fn = const_num_candidates(123)
    assert fn(num_dim=10, num_arms=4) == 123
    with pytest.raises(ValueError, match="must be > 0"):
        const_num_candidates(0)


def test_candidate_gen_config_validation():
    config = CandidateGenConfig()
    assert config.candidate_rv == CandidateRV.SOBOL
    with pytest.raises(ValueError, match="CandidateRV enum"):
        CandidateGenConfig(candidate_rv="SOBOL")
    with pytest.raises(ValueError, match="must be callable"):
        CandidateGenConfig(num_candidates=100)



from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import numpy as np
    from numpy.random import Generator


@dataclass
class NoIncumbentSelector:
    def select(
        self,
        y_obs: np.ndarray,
        mu_obs: np.ndarray | None,
        rng: Generator,
    ) -> int:
        return 0

    def reset(self, rng: Generator) -> None:
        pass



from .draw_acquisition_config import DrawAcquisitionConfig
from .hnr_optimizer_config import HnROptimizerConfig
from .nds_optimizer_config import NDSOptimizerConfig
from .pareto_acquisition_config import ParetoAcquisitionConfig
from .raasp_optimizer_config import RAASPOptimizerConfig
from .random_acquisition_config import RandomAcquisitionConfig
from .ucb_acquisition_config import UCBAcquisitionConfig

AcquisitionConfig = (
    UCBAcquisitionConfig
    | DrawAcquisitionConfig
    | ParetoAcquisitionConfig
    | RandomAcquisitionConfig
)
AcqOptimizerConfig = RAASPOptimizerConfig | HnROptimizerConfig | NDSOptimizerConfig
__all__ = [
    "AcqOptimizerConfig",
    "AcquisitionConfig",
    "DrawAcquisitionConfig",
    "HnROptimizerConfig",
    "NDSOptimizerConfig",
    "ParetoAcquisitionConfig",
    "RAASPOptimizerConfig",
    "RandomAcquisitionConfig",
    "UCBAcquisitionConfig",
]



from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import numpy as np


@dataclass
class PrepareAskResult:
    model: Any
    y_mean: float | None
    y_std: float | None
    lengthscales: np.ndarray | None



from __future__ import annotations
from .turbo_gp_base import TurboGPBase


class TurboGP(TurboGPBase):
    def __init__(
        self,
        train_x,
        train_y,
        likelihood,
        lengthscale_constraint,
        outputscale_constraint,
        ard_dims: int,
    ) -> None:
        import torch
        from gpytorch.kernels import MaternKernel, ScaleKernel
        from gpytorch.means import ConstantMean

        super().__init__(train_x, train_y, likelihood)
        batch_shape = (
            torch.Size(train_y.shape[:-1])
            if getattr(train_y, "ndim", 0) > 1
            else torch.Size()
        )
        self.mean_module = ConstantMean(batch_shape=batch_shape)
        base_kernel = MaternKernel(
            nu=2.5,
            ard_num_dims=ard_dims,
            batch_shape=batch_shape,
            lengthscale_constraint=lengthscale_constraint,
        )
        self.covar_module = ScaleKernel(
            base_kernel,
            batch_shape=batch_shape,
            outputscale_constraint=outputscale_constraint,
        )



import numpy as np
from numpy.random import Generator
from .ackley_core import ackley_core


class Ackley:
    def __init__(self, noise: float, rng: Generator):
        self.noise = noise
        self.rng = rng
        self.bounds = [-32.768, 32.768]

    def __call__(self, x: np.ndarray) -> np.ndarray:
        x = np.asarray(x, dtype=float)
        if x.ndim == 1:
            x = x[None, :]
        y = -ackley_core(x) + self.noise * self.rng.normal(size=(x.shape[0],))
        return y if y.ndim > 0 else float(y)



from __future__ import annotations
from dataclasses import dataclass, field
from .candidate_rv import CandidateRV
from .raasp_driver import RAASPDriver
from .num_candidates_fn import NumCandidatesFn, default_num_candidates


@dataclass(frozen=True)
class CandidateGenConfig:
    candidate_rv: CandidateRV = CandidateRV.SOBOL
    num_candidates: NumCandidatesFn = field(
        default_factory=lambda: default_num_candidates
    )
    raasp_driver: RAASPDriver = RAASPDriver.ORIG

    def __post_init__(self) -> None:
        if not isinstance(self.candidate_rv, CandidateRV):
            raise ValueError(
                f"candidate_rv must be a CandidateRV enum, got {self.candidate_rv!r}"
            )
        if not callable(self.num_candidates):
            raise ValueError(
                f"num_candidates must be callable, got {type(self.num_candidates)!r}"
            )
        test_n = int(self.num_candidates(num_dim=1, num_arms=1))
        if test_n <= 0:
            raise ValueError(f"num_candidates must be > 0, got {test_n}")



from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..components.protocols import AcquisitionOptimizer


@dataclass(frozen=True)
class DrawAcquisitionConfig:
    def build(self) -> AcquisitionOptimizer:
        from ..components.builder import build_acquisition_optimizer

        return build_acquisition_optimizer(self)



from __future__ import annotations

import numpy as np


class AppendableArray:
    def __init__(self, initial_capacity: int = 100) -> None:
        self._initial_capacity = initial_capacity
        self._buffer: np.ndarray | None = None
        self._size = 0
        self._num_cols: int | None = None

    @property
    def shape(self) -> tuple[int, int]:
        if self._num_cols is None:
            return (0, 0)
        return (self._size, self._num_cols)

    def _initialize_buffer(self, row: np.ndarray) -> None:
        if row.ndim == 0:
            self._num_cols = 1
            row = row.reshape(1, 1)
        elif row.ndim == 1:
            self._num_cols = row.shape[0]
            row = row[np.newaxis, :]
        elif row.ndim == 2:
            if row.shape[0] != 1:
                raise ValueError(f"Expected row shape (1, D), got {row.shape}")
            self._num_cols = row.shape[1]
        else:
            raise ValueError(f"Expected 0D, 1D or 2D array, got {row.ndim}D")

        self._buffer = np.empty(
            (self._initial_capacity, self._num_cols), dtype=row.dtype
        )

    def _validate_row(self, row: np.ndarray) -> np.ndarray:
        if row.ndim == 0:
            if self._num_cols != 1:
                raise ValueError(f"Expected {self._num_cols} columns, got 1 (scalar)")
            return row.reshape(1, 1)
        if row.ndim == 1:
            if row.shape[0] != self._num_cols:
                raise ValueError(
                    f"Expected {self._num_cols} columns, got {row.shape[0]}"
                )
            return row[np.newaxis, :]
        if row.ndim == 2:
            if row.shape != (1, self._num_cols):
                raise ValueError(
                    f"Expected shape (1, {self._num_cols}), got {row.shape}"
                )
            return row
        raise ValueError(f"Expected 0D, 1D or 2D array, got {row.ndim}D")

    def append(self, row: np.ndarray) -> None:
        row = np.asarray(row)

        if self._num_cols is None:
            self._initialize_buffer(row)

        row = self._validate_row(row)

        assert self._buffer is not None
        if self._size + 1 > self._buffer.shape[0]:
            new_capacity = self._buffer.shape[0] * 2
            new_buffer = np.empty(
                (new_capacity, self._num_cols), dtype=self._buffer.dtype
            )
            new_buffer[: self._size] = self._buffer[: self._size]
            self._buffer = new_buffer

        self._buffer[self._size] = row
        self._size += 1

    def view(self) -> np.ndarray:
        if self._buffer is None:
            return np.empty((0, 0))
        return self._buffer[: self._size]

    def __len__(self) -> int:
        return self._size

    def __getitem__(self, key) -> np.ndarray:
        return self.view()[key]
